print("import: 'creme'")
import creme

print("import: 'creme.anomaly'")
import creme.anomaly

print("import: 'creme.cluster'")
import creme.cluster

print("import: 'creme.compat'")
import creme.compat

print("import: 'creme.compose'")
import creme.compose

print("import: 'creme.decomposition'")
import creme.decomposition

print("import: 'creme.ensemble'")
import creme.ensemble

print("import: 'creme.feature_extraction'")
import creme.feature_extraction

print("import: 'creme.feature_selection'")
import creme.feature_selection

print("import: 'creme.impute'")
import creme.impute

print("import: 'creme.linear_model'")
import creme.linear_model

print("import: 'creme.metrics'")
import creme.metrics

print("import: 'creme.multiclass'")
import creme.multiclass

print("import: 'creme.multioutput'")
import creme.multioutput

print("import: 'creme.naive_bayes'")
import creme.naive_bayes

print("import: 'creme.neighbors'")
import creme.neighbors

print("import: 'creme.optim'")
import creme.optim

print("import: 'creme.optim.losses'")
import creme.optim.losses

print("import: 'creme.preprocessing'")
import creme.preprocessing

print("import: 'creme.proba'")
import creme.proba

print("import: 'creme.reco'")
import creme.reco

print("import: 'creme.stats'")
import creme.stats

print("import: 'creme.tree'")
import creme.tree

print("import: 'creme.utils'")
import creme.utils

